#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "dlmalloc.h"

word ChangeDir(char *path);
void saveNamedPlaylist(bool photo, char *name);
void saveNamedPlaylistFromKnown(bool photo, char *name, char *dir);
void removeAllFromPlaylist(bool photo);
bool isJpg(char *name);
bool isMp3(char *name);

extern dword write_log;
extern int verbose;			
extern char log_buffer[512];
extern char curr_dir[256];
extern char file_name[100];
extern char from_dir[256];
extern char tracked_dir[256];
extern char known_dir[256];
extern int no_of_playlist_photos;
extern int no_of_playlist_mp3s;
extern bool mp3_playlist_playing;
extern dword recurse_subfolders;

bool ChangeDirPath(char *path);
void clearHistory(bool photo);
//char* GetCurrentDir(void);

#define X	100
#define Y	 80
#define W	520
#define H	500

extern struct playlist
{
	char directory[256];
	char file_name[100];
} *photo_playlist[512], *mp3_playlist[512];

extern int mp3_history[512];
extern word mp3_history_pointer;

/* load the playlist in from file, either photo.pl or mp3.pl or some user named variant */
void loadNamedPlaylist(bool photo, char *name)
{
	word i;
	char pl_name[100], *load_dump;
	TYPE_File *playlist_file;

	// remove any old memory still there
	removeAllFromPlaylist(photo);
	
	ChangeDirPlaylists();

	strcpy(pl_name, name);
	if (photo)
		no_of_playlist_photos = 0;
	else
		no_of_playlist_mp3s = 0;

	if (TAP_Hdd_Exist(pl_name))
	{
		if (write_log) if (verbose) logText("Reading playlist from file", FALSE);
		playlist_file = TAP_Hdd_Fopen(pl_name);
		if (playlist_file != NULL)
		{
			if (photo) {
				TAP_Hdd_Fread(&no_of_playlist_photos, sizeof(int), 1, playlist_file);

				if (no_of_playlist_photos > 0) {
					load_dump = dlmalloc(sizeof(*photo_playlist[0]) * no_of_playlist_photos);
					TAP_Hdd_Fread(load_dump, sizeof(*photo_playlist[0]), no_of_playlist_photos, playlist_file);
					for (i = 0; i < no_of_playlist_photos; i++) {
						photo_playlist[i] = dlmalloc(sizeof(*photo_playlist[i]));
						memcpy(photo_playlist[i], load_dump, sizeof(*photo_playlist[i]));
						load_dump += sizeof(*photo_playlist[i]);
					}

					load_dump -= sizeof(*photo_playlist[0]) * no_of_playlist_photos;
					dlfree(load_dump);
					dlmalloc_trim(0);
				}
			}
			else {
				TAP_Hdd_Fread(&no_of_playlist_mp3s, sizeof(int), 1, playlist_file);

				if (no_of_playlist_mp3s > 0) {
					load_dump = dlmalloc(sizeof(*mp3_playlist[0]) * no_of_playlist_mp3s);
					TAP_Hdd_Fread(load_dump, sizeof(*mp3_playlist[0]), no_of_playlist_mp3s, playlist_file);
					for (i = 0; i < no_of_playlist_mp3s; i++) {
						mp3_playlist[i] = dlmalloc(sizeof(*mp3_playlist[i]));
						memcpy(mp3_playlist[i], load_dump, sizeof(*mp3_playlist[i]));
						load_dump += sizeof(*mp3_playlist[i]);
					}

					load_dump -= sizeof(*mp3_playlist[0]) * no_of_playlist_mp3s;
					dlfree(load_dump);
					dlmalloc_trim(0);
				}
			}
			TAP_Hdd_Fclose(playlist_file);
		}
	}

	ChangeDirPath(curr_dir);
}

void deletePlaylistHistory()
{
	// delete history
	ChangeDirPlaylists();
	if (TAP_Hdd_Exist("mp3.pl.history"))
		TAP_Hdd_Delete("mp3.pl.history");
	ChangeDirPath(curr_dir);
	mp3_playlist_playing = FALSE;
}

/* load the playlist history in from file */
void loadPlaylistHistory()
{
	word i;
	int *load_dump;
	TYPE_File *playlist_file;
	
	ChangeDirPlaylists();
	clearHistory(FALSE);

	if (TAP_Hdd_Exist("mp3.pl.history"))	{
		if (write_log) if (verbose) logText("Reading playlist history from file", FALSE);
		playlist_file = TAP_Hdd_Fopen("mp3.pl.history");
		if (playlist_file != NULL) {
			load_dump = dlmalloc(sizeof(int) * 512);
			TAP_Hdd_Fread(&mp3_history_pointer, sizeof(int), 1, playlist_file);
			TAP_Hdd_Fread(load_dump, sizeof(int), 512, playlist_file);
			for (i = 0; i < 512; i++) {
				mp3_history[i] = *load_dump;
				load_dump++;
			}

			load_dump -= 512;
			dlfree(load_dump);
			dlmalloc_trim(0);

			TAP_Hdd_Fclose(playlist_file);
		}
	}

	ChangeDirPath(curr_dir);
}

void appendToPlaylist(bool photo, char *directory, char *file_name, bool save_to_disk)
{
	if (photo) {
		if (no_of_playlist_photos >= 512)
			return ;
	
		photo_playlist[no_of_playlist_photos] = dlmalloc(sizeof(*photo_playlist[no_of_playlist_photos]));
		strcpy(photo_playlist[no_of_playlist_photos]->directory, directory);
		strcpy(photo_playlist[no_of_playlist_photos]->file_name, file_name);
		no_of_playlist_photos++;
		if (save_to_disk) saveNamedPlaylistFromKnown(TRUE, "photo.pl", directory);
	}
	else {
		if (no_of_playlist_mp3s >= 512)
			return ;

		mp3_playlist[no_of_playlist_mp3s] = dlmalloc(sizeof(*mp3_playlist[no_of_playlist_mp3s]));
		strcpy(mp3_playlist[no_of_playlist_mp3s]->directory, directory);
		strcpy(mp3_playlist[no_of_playlist_mp3s]->file_name, file_name);
		no_of_playlist_mp3s++;
		if (save_to_disk) saveNamedPlaylistFromKnown(FALSE, "mp3.pl", directory);
	}
}

void appendRecursivelyToPlaylist(bool photo)
{
	TYPE_File file;
	char temp[256];
	char **subfolders = dlmalloc(sizeof(*subfolders) * 1024);
	int subfolder_count = 0, i;

	i = TAP_Hdd_FindFirst(&file);

	while(i-- && ((photo && no_of_playlist_photos < 512) || (!photo && no_of_playlist_mp3s < 512)))
	{
		if (photo && isJpg(file.name) && existsInPlaylist(tracked_dir, file.name) < 0)
		{
			photo_playlist[no_of_playlist_photos] = dlmalloc(sizeof(*photo_playlist[no_of_playlist_photos]));
			strcpy(photo_playlist[no_of_playlist_photos]->directory, tracked_dir);
			strcpy(photo_playlist[no_of_playlist_photos]->file_name, file.name);
			no_of_playlist_photos++;
		}
		ef (!photo && isMp3(file.name) && existsInPlaylist(tracked_dir, file.name) < 0)
		{
			mp3_playlist[no_of_playlist_mp3s] = dlmalloc(sizeof(*mp3_playlist[no_of_playlist_mp3s]));
			strcpy(mp3_playlist[no_of_playlist_mp3s]->directory, tracked_dir);
			strcpy(mp3_playlist[no_of_playlist_mp3s]->file_name, file.name);
			no_of_playlist_mp3s++;
		}
		ef (file.attr == ATTR_FOLDER)	
		{
			if (recurse_subfolders)
			{
				subfolders[subfolder_count] = dlmalloc(100 * sizeof(char));
				strcpy(subfolders[subfolder_count], file.name);
				subfolder_count++;
			}
		}

		if (i == 0 || TAP_Hdd_FindNext(&file) == 0)
			break;
	}

	if (recurse_subfolders)
	{
		for (i = 0; i < subfolder_count; i++)
		{
			ChangeDir(subfolders[i]);
			appendRecursivelyToPlaylist(photo);
			ChangeDir("..");
		}

		for (i = 0; i < subfolder_count; i++)
		{
			dlfree(subfolders[i]);
			subfolders[i] = NULL;
		}
		dlfree(subfolders);
		subfolders = NULL;
	}

	dlmalloc_trim(0);
}

void appendDirectoryToPlaylist(char *directory)
{
	bool photo = !InMp3FilesFolder(curr_dir) && (strcmp(directory, "MP3") != 0);

	ChangeDir(directory);
	appendRecursivelyToPlaylist(photo);
	
	if (photo) {
		saveNamedPlaylist(TRUE, "photo.pl");
	}
	else {
		saveNamedPlaylist(FALSE, "mp3.pl");
	}
	
	ChangeDirPath(curr_dir);
}

void removeFromPlaylist(bool photo, int delete_idx)
{
	int i;

	if (photo) {
		// shuffle up playlist files and delete last entry
		for (i = delete_idx; i < (no_of_playlist_photos-1); i++)
		{
			strcpy(photo_playlist[i]->directory, photo_playlist[i+1]->directory);
			strcpy(photo_playlist[i]->file_name, photo_playlist[i+1]->file_name);
		}
		dlfree(photo_playlist[no_of_playlist_photos-1]);
		no_of_playlist_photos--;
	}
	else {
		// shuffle up remaining playlist files
		for (i = delete_idx; i < (no_of_playlist_mp3s-1); i++)
		{
			strcpy(mp3_playlist[i]->directory, mp3_playlist[i+1]->directory);
			strcpy(mp3_playlist[i]->file_name, mp3_playlist[i+1]->file_name);
		}
		dlfree(mp3_playlist[no_of_playlist_mp3s-1]);
		no_of_playlist_mp3s--;
	}

	dlmalloc_trim(0);

	if (photo) {
		saveNamedPlaylist(TRUE, "photo.pl");
	}
	else {
		saveNamedPlaylist(FALSE, "mp3.pl");
		deletePlaylistHistory();
	}
}

void doNamedPlaylistTidy(bool photo, char * name)
{
	int i;
	//char *current_folder;

	loadNamedPlaylist(photo, name);

	// check entries starting at the end
	if (photo) {
		if (no_of_playlist_photos == 0)
			return;

		for (i = no_of_playlist_photos-1; i >= 0; i--)
		{
			ChangeDirPath(photo_playlist[i]->directory);
			if (strcmp(photo_playlist[i]->directory, tracked_dir) != 0) {
				removeFromPlaylist(photo, i);
			}
			else {
				if (!TAP_Hdd_Exist(photo_playlist[i]->file_name)) {
					removeFromPlaylist(photo, i);
				}
			}
		}
	}
	else {
		if (no_of_playlist_mp3s == 0)
			return;

		for (i = no_of_playlist_mp3s-1; i >= 0; i--)
		{
			ChangeDirPath(mp3_playlist[i]->directory);
			if (strcmp(mp3_playlist[i]->directory, tracked_dir) != 0) {
				removeFromPlaylist(photo, i);
			}
			else {
				if (!TAP_Hdd_Exist(mp3_playlist[i]->file_name)) {
					removeFromPlaylist(photo, i);
				}
			}
		}
	}

	dlmalloc_trim(0);
	saveNamedPlaylist(photo, name);
}

void movePlalistEntryDown(bool photo, int shuffle_idx)
{
	if (photo) {
		if (shuffle_idx >= no_of_playlist_photos - 1)
			return;

		// swap playlist entry with one above
		strcpy(from_dir, photo_playlist[shuffle_idx+1]->directory);
		strcpy(file_name, photo_playlist[shuffle_idx+1]->file_name);
		strcpy(photo_playlist[shuffle_idx+1]->directory, photo_playlist[shuffle_idx]->directory);
		strcpy(photo_playlist[shuffle_idx+1]->file_name, photo_playlist[shuffle_idx]->file_name);
		strcpy(photo_playlist[shuffle_idx]->directory, from_dir);
		strcpy(photo_playlist[shuffle_idx]->file_name, file_name);
	}
	else {
		if (shuffle_idx >= no_of_playlist_mp3s - 1)
			return;

		// swap playlist entry with one above
		strcpy(from_dir, mp3_playlist[shuffle_idx+1]->directory);
		strcpy(file_name, mp3_playlist[shuffle_idx+1]->file_name);
		strcpy(mp3_playlist[shuffle_idx+1]->directory, mp3_playlist[shuffle_idx]->directory);
		strcpy(mp3_playlist[shuffle_idx+1]->file_name, mp3_playlist[shuffle_idx]->file_name);
		strcpy(mp3_playlist[shuffle_idx]->directory, from_dir);
		strcpy(mp3_playlist[shuffle_idx]->file_name, file_name);
		deletePlaylistHistory();
	}
}

void movePlalistEntryUp(bool photo, int shuffle_idx)
{
	if (photo) {
		if (shuffle_idx <= 0)
			return;

		// swap playlist entry with one above
		strcpy(from_dir, photo_playlist[shuffle_idx-1]->directory);
		strcpy(file_name, photo_playlist[shuffle_idx-1]->file_name);
		strcpy(photo_playlist[shuffle_idx-1]->directory, photo_playlist[shuffle_idx]->directory);
		strcpy(photo_playlist[shuffle_idx-1]->file_name, photo_playlist[shuffle_idx]->file_name);
		strcpy(photo_playlist[shuffle_idx]->directory, from_dir);
		strcpy(photo_playlist[shuffle_idx]->file_name, file_name);
	}
	else {
		if (shuffle_idx <= 0)
			return;

		// swap playlist entry with one above
		strcpy(from_dir, mp3_playlist[shuffle_idx-1]->directory);
		strcpy(file_name, mp3_playlist[shuffle_idx-1]->file_name);
		strcpy(mp3_playlist[shuffle_idx-1]->directory, mp3_playlist[shuffle_idx]->directory);
		strcpy(mp3_playlist[shuffle_idx-1]->file_name, mp3_playlist[shuffle_idx]->file_name);
		strcpy(mp3_playlist[shuffle_idx]->directory, from_dir);
		strcpy(mp3_playlist[shuffle_idx]->file_name, file_name);
		deletePlaylistHistory();
	}
}

void removeAllFromPlaylist(bool photo)
{
	int i;

	if (photo) {
		for (i = 0; i < no_of_playlist_photos; i++)
		{
			dlfree(photo_playlist[i]);
		}
		no_of_playlist_photos = 0;
	}
	else {
		for (i = 0; i < no_of_playlist_mp3s; i++)
		{
			dlfree(mp3_playlist[i]);
		}
		no_of_playlist_mp3s = 0;
		deletePlaylistHistory();
	}

	dlmalloc_trim(0);
}

int existsInPlaylist(char *directory, char *file_name)
{
	int i;

	if (isJpg(file_name)) {
		for (i = 0; i < no_of_playlist_photos; i++)
		{
			if (strcmp(directory, photo_playlist[i]->directory) == 0 && strcmp(file_name, photo_playlist[i]->file_name) == 0)
				return i;
		}
	}
	ef (isMp3(file_name)) {
		for (i = 0; i < no_of_playlist_mp3s; i++)
		{
			if (strcmp(directory, mp3_playlist[i]->directory) == 0 && strcmp(file_name, mp3_playlist[i]->file_name) == 0)
				return i;
		}
	}

	return -1;
}

void savePlaylistDetails(bool photo, char *name)
{
	word i;
	char *save_dump;
	TYPE_File *playlist_file;

	if (photo) {
		TAP_Hdd_Delete(name);
		TAP_Hdd_Create(name, ATTR_NORMAL);
		playlist_file = TAP_Hdd_Fopen(name);
		TAP_Hdd_Fwrite(&no_of_playlist_photos, sizeof(int), 1, playlist_file);

		if (no_of_playlist_photos > 0) {
			save_dump = dlmalloc(sizeof(*photo_playlist[0]) * no_of_playlist_photos);

			for (i = 0; i < no_of_playlist_photos; i++) {
				memcpy(save_dump, photo_playlist[i], sizeof(*photo_playlist[i]));
				save_dump += sizeof(*photo_playlist[i]);
			}
			save_dump -= sizeof(*photo_playlist[0]) * no_of_playlist_photos;
			TAP_Hdd_Fwrite(save_dump, sizeof(*photo_playlist[i]), no_of_playlist_photos, playlist_file);
			
			dlfree(save_dump);
			dlmalloc_trim(0);
		}

		//for (i = 0; i < no_of_playlist_photos; i++)
		//	TAP_Hdd_Fwrite(photo_playlist[i], sizeof(*photo_playlist[i]), 1, playlist_file);

		TAP_Hdd_Fclose(playlist_file);
	}
	else {
		TAP_Hdd_Delete(name);
		TAP_Hdd_Create(name, ATTR_NORMAL);
		playlist_file = TAP_Hdd_Fopen(name);
		TAP_Hdd_Fwrite(&no_of_playlist_mp3s, sizeof(int), 1, playlist_file);

		if (no_of_playlist_mp3s > 0) {
			save_dump = dlmalloc(sizeof(*mp3_playlist[0]) * no_of_playlist_mp3s);

			for (i = 0; i < no_of_playlist_mp3s; i++) {
				memcpy(save_dump, mp3_playlist[i], sizeof(*mp3_playlist[i]));
				save_dump += sizeof(*mp3_playlist[i]);
			}
			save_dump -= sizeof(*mp3_playlist[0]) * no_of_playlist_mp3s;
			TAP_Hdd_Fwrite(save_dump, sizeof(*mp3_playlist[i]), no_of_playlist_mp3s, playlist_file);
			
			dlfree(save_dump);
			dlmalloc_trim(0);
		}

		//for (i = 0; i < no_of_playlist_mp3s; i++)
		//	TAP_Hdd_Fwrite(mp3_playlist[i], sizeof(*mp3_playlist[i]), 1, playlist_file);

		TAP_Hdd_Fclose(playlist_file);
	}
}

void saveNamedPlaylist(bool photo, char *name)
{
	// save playlist to file
	ChangeDirPlaylists();
	savePlaylistDetails(photo, name);
	ChangeDirPath(curr_dir);
}

void saveNamedPlaylistFromKnown(bool photo, char *name, char *dir)
{
	strcpy(known_dir, dir);

	// save playlist list to file
	ChangeDirPlaylists();
	savePlaylistDetails(photo, name);
	ChangeDirPath(known_dir);
}

void savePlaylistHistory()
{
	word i;
	int *save_dump;
	TYPE_File *playlist_file;

	ChangeDirPlaylists();

	TAP_Hdd_Delete("mp3.pl.history");
	TAP_Hdd_Create("mp3.pl.history", ATTR_NORMAL);
	playlist_file = TAP_Hdd_Fopen("mp3.pl.history");
	TAP_Hdd_Fwrite(&mp3_history_pointer, sizeof(int), 1, playlist_file);
	save_dump = dlmalloc(sizeof(int) * 512);
	for (i = 0; i < 512; i++) {
		if (mp3_history[i] >= 0)
		*save_dump = mp3_history[i];
		save_dump++;
	}
	save_dump -= 512;
	TAP_Hdd_Fwrite(save_dump, sizeof(int), 512, playlist_file);

	dlfree(save_dump);
	dlmalloc_trim(0);
	TAP_Hdd_Fclose(playlist_file);

	ChangeDirPath(curr_dir);
}