#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

/*
 **  id3.c
 **  $Id: id3.c,v 1.22 2005/03/28 23:25:03 brian Exp $
 **  This is something I really did not want to write.
 **
 **  It was pretty annoying that I had to write this
 **  but I could not find one single ID3 library
 **  that had a BSD license and was written in C.
 */ 

/* ====================================================================
 * Copyright (c) 2003 Brian Aker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * 4. The name "libID3" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    brian@tangent.org.
 *
 * 5. Products derived from this software may not be called "libID3".
 *
 * 6. For any changes to the license please contact brian@tangent.org.
 *    He is quite willing to be flexible for you.
 *
 * THIS SOFTWARE IS PROVIDED BY BRIAN AKER ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL BRIAN AKER OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include "tap.h"
#include "id3.h"
//#include <config.h>
//#include "md5.h"
#include "dlmalloc.h"

//void md5_signature(const byte *buf, int length, char *result);

#ifdef _MSC_VER
// #define snprintf _snprintf
#endif /* _MSC_VER */

const char id3_genres[GENRE_MAX][25] =
{
  "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge", "Hip-Hop",
  "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B", "Rap", "Reggae",
  "Rock", "Techno", "Industrial", "Alternative", "Ska", "Death Metal", "Pranks", "Soundtrack",
  "Euro-Techno", "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical",
  "Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise", "Alt",
  "Bass", "Soul", "Punk", "Space", "Meditative", "Instrumental Pop",
  "Instrumental Rock", "Ethnic", "Gothic", "Darkwave", "Techno-Industrial", "Electronic",
  "Pop-Folk", "Eurodance", "Dream", "Southern Rock", "Comedy", "Cult",
  "Gangsta Rap", "Top 40", "Christian Rap", "Pop/Funk", "Jungle", "Native American",
  "Cabaret", "New Wave", "Psychedelic", "Rave", "Showtunes", "Trailer", "Lo-Fi", "Tribal",
  "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll", "Hard Rock", "Folk",
  "Folk/Rock", "National Folk", "Swing", "Fast-Fusion", "Bebob", "Latin", "Revival",
  "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock", "Progressive Rock",
  "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band", "Chorus", "Easy Listening",
  "Acoustic", "Humour", "Speech", "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony",
  "Booty Bass", "Primus", "Porn Groove", "Satire", "Slow Jam", "Club", "Tango",
  "Samba", "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle", "Duet",
  "Punk Rock", "Drum Solo", "A Cappella", "Euro-House", "Dance Hall", "Goa",
  "Drum & Bass", "Club-House", "Hardcore", "Terror", "Indie", "BritPop", "Negerpunk",
  "Polsk Punk", "Beat", "Christian Gangsta Rap", "Heavy Metal", "Black Metal", "Crossover",
  "Contemporary Christian", "Christian Rock", "Merengue", "Salsa", "Thrash Metal",
  "Anime", "JPop", "Synthpop"
};

const char * genre_string(int genre) 
{
  if (genre < GENRE_MAX && genre > -1)
    return id3_genres[genre];

  return NULL;
}

void clean_string(byte *string, int length) 
{
  byte *ptr= string;

  for(; (ptr - string) < length; ptr++) {
    if (!isprint(*ptr) && (*ptr < 192 || *ptr > 252)) 
       *ptr= ' ';
  }
}

char * add_tag(ID3 *info, const byte *tag, size_t length) 
{
  char *begin;

  // If the current tag is too big for our block of memory
  if ((length + info->length) > info->size) 
  {
    if ( info->user_memory) 
    {
      // This because we didn't allocate the memory in the first place, and this should be an error
      return NULL;
    }
    else if ( info->size == 0 ) 
    {
      info->buffer= (char *) dlmalloc(ID3_INIT_SIZE);
      info->size= ID3_INIT_SIZE;
      info->ptr= info->buffer;
    }
    else if ( length > info->tag_length  ) 
    {
      info->buffer= dlrealloc(info->buffer, info->tag_length);
      info->ptr= info->buffer + info->length;
      info->size= info->tag_length;
    }
    else
    {
      /*
        This should never, ever happen
        since we don't support compressed tags just yet
      */
      return NULL;
    }
  }
  begin= info->ptr;
  memcpy(info->ptr, tag, length);	
  clean_string(info->ptr, length);
  info->length += length + 1;
  info->ptr += length;
  *info->ptr++= '\0';

  // Now return the position of the last string
  return begin;
}

void select_tag(ID3 *info, const char *name, const char *ptr, size_t length) 
{
  if (!strcmp(name, "title"))
  {
    info->title= add_tag(info, ptr, length);
    info->title_size= length;
  }
  else if (!strcmp(name, "artist"))
  {
    info->artist= add_tag(info, ptr, length);
    info->artist_size= length;
  }
  else if (!strcmp(name, "album"))
  {
    info->album= add_tag(info, ptr, length);
    info->album_size= length;
  }
  else if (!strcmp(name, "year"))
  {
    info->year= add_tag(info, ptr, length);
    info->year_size= length;
  }
  else if (!strcmp(name, "comment"))
  {
    info->comment= add_tag(info, ptr, length);
    info->comment_size= length;
  }
  else if (!strcmp(name, "track"))
  {
    info->track= add_tag(info, ptr, length);
	info->track_number = atoi(info->track);
    info->track_size= length;
  }
  else if (!strcmp(name, "length"))
  {
    info->length_millis= add_tag(info, ptr, length);
	info->duration = atoi(info->length_millis) / 1000;
    info->length_millis_size= length;
  }
  else if (!strcmp(name, "picture"))
  {
    word i = 0; const byte *pic = ptr;

	if (length == 0) return;

	pic++; i++;					// encoding
	if (i >= length) return;
	while (*pic != '\0') {		// mime type
		pic++; i++;
		if (i >= length) return;
	}
	pic++; i++;					
	if (i >= length) return;
	pic++; i++;					// pic type
	if (i >= length) return;
	while (*pic != '\0') {		// desc
		pic++; i++;
		if (i >= length) return;
	}
	pic++; i++;
	if (i >= length) return;

	if (i+2 < length && pic[0] == 0xff && pic[1] == 0xd8) {
		if (info->picture) dlfree(info->picture);
		info->picture = dlmalloc(length - i);
		memcpy(info->picture, pic, length - i); 
		info->picture_size = length - i;
	}
  }
  else if (!strcmp(name, "genre"))
  {
	word i;

    info->genre_code= add_tag(info, ptr, length);
	
	strcpy(info->genre, info->genre_code);
	for (i = 0; i < length; i++) {
		if (info->genre_code[i] == '(' && i < length-2) {
			strcpy(info->genre, genre_string(atoi(info->genre_code+i+1)));
			break;
		}
	}

	info->genre_size= strlen(info->genre);
  }
  else if (!strcmp(name, "encoder"))
  {
    info->encoder= add_tag(info, ptr, length);
    info->encoder_size= length;
  }
  else if (!strcmp(name, "bpm"))
  {
    info->encoder= add_tag(info, ptr, length);
    info->encoder_size= length;
  }
}


size_t id3_size(const byte* buffer) 
{
  size_t size = 0;
  size=  (buffer[3] & 0x7f) + 
    ((buffer[2] & 0x7f) << 7) +
    ((buffer[1] & 0x7f) << 14) +
    ((buffer[0] & 0x7f) << 21);

  return size;
}

size_t id3_size2(const byte * buffer) 
{
  size_t size = 0;
  size=  (buffer[2] & 0x7f) + 
    ((buffer[1] & 0x7f) << 7) +
    ((buffer[0] & 0x7f) << 14);

  return size;
}

size_t get_framesize(const byte *buffer) 
{
  return ((buffer[5] << 16) + (buffer[6] << 8) + (buffer[7]));
}

int get_id3v1_tag (ID3 *info, byte *blob, size_t blob_length) 
{
  byte *ptr = blob;
  const char *genre = NULL;

  ptr += (blob_length - 128);

  if (!memcmp(ptr, "TAG", 3)) 
  {
    strcpy(info->id3v1_version, "1");

    /* Paranoid, not all systems are made equally */
    ptr +=3;

	if (info->mask & TITLE_TAG) 
      info->processor(info, "title", ptr, 30);
    ptr +=30;

    if (info->mask & ARTIST_TAG)
      info->processor(info, "artist", ptr, 30);
    ptr +=30;

    if (info->mask & ALBUM_TAG)
      info->processor(info, "album", ptr, 30);
    ptr +=30;

    if (info->mask & YEAR_TAG)
      info->processor(info, "year", ptr, 4);
    ptr +=4;

	if (info->mask & COMMENT_TAG) {
      info->processor(info, "comment", ptr, 30);
	  if (ptr[28] == '\0' && ptr[29] != '\0') {
		  strcpy(info->id3v1_version, "1.1");
		  info->track_number = ptr[29];
	  }
	}
    ptr +=30;

	if (info->mask & GENRE_TAG) {
      genre= genre_string(ptr[0]);
	}
    if (genre)
      info->processor(info, "genre", genre, strlen(genre));

    return 0;
  }

  return ID3_ERR_NO_TAG;
}

int id_2_2(ID3 *info, byte *blob) 
{
  unsigned char *ptr= blob;

  while (info->tag_length > (ptr - blob)) 
  {
    size_t size= id3_size2(ptr+3);

    if (size <= 0)
    {
      return 0;
    }

    if (size + (ptr - blob) > info->tag_length) 
      break;

    if (!strncmp("TP1", ptr, 3) && (info->mask & ARTIST_TAG))
    {
      info->processor(info, "artist", ptr + 6, size);
      info->mask_found |= ARTIST_TAG;
    }
    else if (!strncmp("TT2", ptr, 3) && (info->mask & TITLE_TAG)) 
    {
      info->processor(info, "title", ptr + 6, size);
      info->mask_found |= TITLE_TAG;
    }
    else if (!strncmp("TAL", ptr, 3) && (info->mask & ALBUM_TAG)) 
    {
      info->processor(info, "album", ptr + 6, size);
      info->mask_found |= ALBUM_TAG;
    }
    else if (!strncmp("TRK", ptr, 3) && (info->mask & TRACK_TAG)) 
    {
      info->processor(info, "track", ptr + 6, size);
      info->mask_found |= TRACK_TAG;
    }
    else if (!strncmp("TEN", ptr, 3) && (info->mask & ENCODER_TAG)) 
    {
      info->processor(info, "encoder", ptr + 6, size);
      info->mask_found |= ENCODER_TAG;
    }
    else if (!strncmp("TYE", ptr, 3) && (info->mask & YEAR_TAG)) 
    {
      info->processor(info, "year", ptr + 6, size);
      info->mask_found |= YEAR_TAG;
    }
    else if (!strncmp("COM", ptr, 3) && (info->mask & COMMENT_TAG)) 
    {
      info->processor(info, "comment", ptr + 6, size);
      info->mask_found |= COMMENT_TAG;
    }
    else if (!strncmp("TCO", ptr, 3) && (info->mask & GENRE_TAG)) 
    {
      info->processor(info, "genre", ptr + 6, size);
      info->mask_found |= GENRE_TAG;
    } 
    else
    {
      char temp[4];
      memset(temp, 0, 4);
      memcpy(temp, ptr, 3);
      info->processor(info, temp, ptr + 6, size);
    }

    if (info->mask_found == info->mask)
      return 0;

    ptr += size + 6;
  }

  return 0;
}

int id_2_3(ID3 *info, byte *blob, bool read_art) 
{
  byte *ptr= blob;

  while (info->tag_length > (ptr - blob)) 
  {
    size_t size= get_framesize(ptr);

    if (size + (ptr - blob) > info->tag_length) 
      break;

    if (size <= 0)
      return 0;

    if (!strncmp("TPE1",ptr,4) && (info->mask & ARTIST_TAG)) 
    {
      info->processor(info, "artist", ptr +10, size);
      info->mask_found |= ARTIST_TAG;
    }
    else if (!strncmp("TIT2",ptr,4) && (info->mask & TITLE_TAG)) 
    {
      info->processor(info, "title", ptr +10, size);
      info->mask_found |= TITLE_TAG;
    }
    else if (!strncmp("TALB",ptr,4) && (info->mask & ALBUM_TAG)) 
    {
      info->processor(info, "album", ptr +10, size);
      info->mask_found |= ARTIST_TAG;
    }
    else if (!strncmp("TRCK",ptr,4) && (info->mask & TRACK_TAG)) 
    {
      info->processor(info, "track", ptr +10, size);
      info->mask_found |= TRACK_TAG;
    }
    else if (!strncmp("TYER",ptr,4) && (info->mask & YEAR_TAG)) 
    {
      info->processor(info, "year", ptr +10, size);
      info->mask_found |= YEAR_TAG;
    }
    else if (!strncmp("TENC",ptr,4) && (info->mask & ENCODER_TAG)) 
    {
      info->processor(info, "encoder", ptr +10, size);
      info->mask_found |= ENCODER_TAG;
    }
    else if (!strncmp("COMM",ptr,4) && (info->mask & COMMENT_TAG)) 
    {
      info->processor(info, "comment", ptr +10, size);
      info->mask_found |= COMMENT_TAG;
    }
	else if (!strncmp("TCON",ptr,4) && (info->mask & GENRE_TAG)) 
    {
      info->processor(info, "genre", ptr +10, size);
      info->mask_found |= GENRE_TAG;
    }
	else if (!strncmp("TLEN",ptr,4) && (info->mask & LENGTH_TAG)) 
    {
      info->processor(info, "length", ptr +10, size);
      info->mask_found |= LENGTH_TAG;
    }
	else if (!strncmp("APIC",ptr,4) && (info->mask & PICTURE_TAG)) 
    {
	  if (read_art) {
		  info->processor(info, "picture", ptr +10, size);
		  info->mask_found |= PICTURE_TAG;
	  }
    }
    else
    {
      char temp[5];

      memset(temp, 0, 5);
      memcpy(temp, ptr, 4);
      info->processor(info, temp, ptr + 10, size);
    }
    if (info->mask_found == info->mask)
      return 0;

    ptr += 10 + size;
  }

  return 0;
}

int process_extended_header(ID3 *info, byte *blob) 
{
  dword CRC= 0;
  dword paddingsize= 0;

  /*
    Shortcut. The extended header size will be either 6 or 10 bytes. 
    If it's ten bytes, it means that there's CRC data (though we check
    the flag anyway). I'm gonna save it, though I'll be damned if I 
    know what to do with it.
  */
  if ((blob[3] == 0x0A) && (blob[4])) 
    CRC= (blob[10] << 24) + (blob[11] << 16) +
      (blob[12] << 8) + (blob[13]);

  paddingsize= (blob[6] << 24) + (blob[7] << 16) +
    (blob[8] << 8) + (blob[9]);

  /*subtract the size of the padding from the size of the tag */
  info->tag_length -= paddingsize;

  /*continue decoding the frames */
  return id_2_3(info, blob, FALSE);
}

int get_id3v2_tag (ID3 *info, byte *blob, size_t blob_length, bool read_art) 
{
  int unsynchronized= 0;
  int hasExtendedHeader= 0;
  byte *ptr_buffer = blob;

  if (ptr_buffer && !strncmp(ptr_buffer, "ID3", 3)) 
  {
    info->tag_length= id3_size(ptr_buffer+6);
    //TAP_SPrint(info->id3v2_version, "2.%d.%d", blob[3], blob[4]);
	TAP_SPrint(info->id3v2_version, "2.%d", blob[3]);

    if ((blob[5] & 0x40) >> 6) 
      hasExtendedHeader= 1;

    if ((blob[5] & 0x80) >> 7) 
      unsynchronized= 1;

    /* We can not handle experimental tags */
    if ((blob[5] & 0x20) >> 5) 
      return 1;

#ifdef NOT_SUPPORTED
    if (unsynchronized) 
    {
      register int i,j; /*idx*/
      /*
        Replace every instance of '0xFF 0x00'
        with '0xFF'
      */
      for(i=10; i < info->tag_length; i++) 
        if (blob[i] == 0xFF && blob[i+1] == 0x00) 
          for(j=i+1; i < info->tag_length; i++) 
            blob[j]= blob[j+1];

    }
#endif

    /* Move past the above */
    ptr_buffer += 10;
    /*If the tag has an extended header, parse it*/
    if (hasExtendedHeader) 
      return process_extended_header(info, blob);
    else if (blob[3] == 2) 
      return id_2_2(info, ptr_buffer);
    else if (blob[3] == 3) 
      return id_2_3(info, ptr_buffer, read_art);
    // Yes, I will need to fix this at some point
    else if (blob[3] == 4) 
      return id_2_3(info, ptr_buffer, read_art);
    else 
      return ID3_ERR_UNSUPPORTED_FORMAT;
  }

  return 1;
}

/*int ID3_to_file(unsigned char *filename, unsigned char *newfile) 
{
  int rc= 0;
  caddr_t blob, tag;
  struct stat blob_buf;
  int fd= -1;
  unsigned char *ptr= NULL;

  size_t tag_length= 0;

  if (stat(filename, &blob_buf)) 
    return -1;


  if ((fd= open(filename, O_RDONLY)) ==  -1) 
    return -1;

  blob= mmap(NULL, blob_buf.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
  if ((blob == NULL) || (blob ==(caddr_t)-1)) 
    return ID3_ERR_EMPTY_FILE;

  close(fd);

  // Minimal size blob
  if (blob_buf.st_size  < 10)
  {
    munmap(blob, blob_buf.st_size);
    return ID3_ERR_NO_TAG;
  }



  if ((blob_buf.st_size > 128)) 
  {
    ptr= blob + (blob_buf.st_size - 128);
    //  Set ptr back to NULL if it is not pointing to a tag
    if (memcmp(ptr, "TAG", 3)) 
      ptr= NULL;
  }

  if (!strncmp(blob, "ID3", 3)) 
  {
    tag_length= id3_size(blob+6);
  }

  if (tag_length || ptr)
  {
    FILE *new= NULL;
    size_t total= tag_length + (ptr ? 128 : 0);

    if(!(new= fopen(newfile,"w+")))
      return -1;
    ftruncate(fileno(new), total);
    tag= mmap(NULL, total, PROT_READ | PROT_WRITE, MAP_SHARED, fileno(new), 0);
    if ((tag == NULL) || (tag ==(caddr_t)-1)) 
      return ID3_ERR_EMPTY_FILE;

    fclose(new);

    if (tag_length)
      memcpy(tag, blob, tag_length);

    if (ptr)
    {
      memcpy(tag + tag_length, ptr, 128);
    }
    munmap(tag, total);
  }

  munmap(blob, blob_buf.st_size);

  return rc;
}*/

bool parse_file_ID3(ID3 *info, byte *filename, bool read_art) 
{
	TYPE_File *mp3_file;
	long size, tag_length = 0;
	bool success = 0;
	byte *id3v1_data = NULL, *id3v2_data = NULL, *frame_data = dlmalloc(8192), header[10];

	mp3_file = TAP_Hdd_Fopen(filename);
	if (!mp3_file)
		return FALSE;

	size = TAP_Hdd_Flen(mp3_file);
	//if ((size > 0) && (size % 512) == 0)
	//	size = size + 512;

	TAP_Hdd_Fread(header, 10, 1, mp3_file);
	if (strncmp(header, "ID3", 3) == 0) {
		tag_length = id3_size(header+6);

		id3v2_data = dlmalloc(tag_length+10);
		TAP_Hdd_Fseek(mp3_file, 0, 0);
		TAP_Hdd_Fread(id3v2_data, tag_length+10, 1, mp3_file);
		TAP_Hdd_Fread(frame_data, 8192, 1, mp3_file);
	}

	id3v1_data = dlmalloc(128);
	TAP_Hdd_Fseek(mp3_file, -128, 2);
	TAP_Hdd_Fread(id3v1_data, 128, 1, mp3_file);

	success = parse_blob_ID3(info, id3v1_data, id3v2_data, 128, tag_length, read_art);

	if (!id3v2_data) {
		TAP_Hdd_Fseek(mp3_file, 0, 0);
		TAP_Hdd_Fread(frame_data, 8192, 1, mp3_file);
	}

	parse_mp3_frames(info, frame_data, size, id3v2_data ? tag_length + 10 : 0);

	TAP_Hdd_Fclose(mp3_file);
	dlfree(id3v1_data);
	if (id3v2_data) dlfree(id3v2_data);
	dlfree(frame_data);
	dlmalloc_trim(0);

	return success;
}

void parse_mp3_frames(ID3 *info, byte *buffer, long file_size, long tag_length)
{
	mp3_header header, next_header;
	bool valid_header = FALSE;
	byte* data = buffer; word pointer = 0;

	while (!valid_header && pointer < 8000) {
		valid_header = parse_mp3_header(data, &header);

		if (valid_header) {
			if (mp3_check_vbr(data, &header)) {
				header.bitrate = 0;
				break;
			}

			valid_header = parse_mp3_header(data + header.frame_length, &next_header);
			if (header.version != next_header.version || header.layer != next_header.layer || header.samplerate != next_header.samplerate)
				valid_header = FALSE;

			if (valid_header && header.bitrate != next_header.bitrate) {
				header.bitrate = 0;
				break;
            }
		}

		data++;
		pointer++;
	}

	if (valid_header) {
		info->mp3_version	= header.version;
		info->mp3_layer		= header.layer;
		info->channels		= header.channels;
		info->bitrate		= header.bitrate;
		info->samplerate	= header.samplerate;
		info->duration		= 0;

		if (header.number_of_frames == 0 && header.bitrate > 0)
			header.number_of_frames = (file_size - tag_length) / header.frame_length;

		if (header.number_of_frames > 0) {
			//info->duration = (number_of_frames * 26) / 1000;
			info->duration = (header.number_of_frames * header.samples_per_frame) / header.samplerate;
		}
	}
}

bool parse_mp3_header(byte *buffer, mp3_header *header)
{
	int bitrate_idx, samplerate_idx, mode, emphasis;

	if (((buffer[0] << 4) | ((buffer[1] >> 4) & 0xE)) != 0xFFE)
		return FALSE;

    switch ((buffer[1] >> 3 & 0x3)) {
	    case 3:
	        header->version = 0;
			break;
		case 2:
			header->version = 1;
			break;
		case 0:
			header->version = 2;
			break;
		default:
			return FALSE;
	}
	bitrate_idx = (buffer[2] >> 4) & 0xF;
	samplerate_idx = (buffer[2] >> 2) & 0x3;
	mode = (buffer[3] >> 6) & 0x3;
	header->layer = 4 - ((buffer[1] >> 1) & 0x3);
	emphasis = (buffer[3]) & 0x3;

	if (bitrate_idx == 0xF || samplerate_idx == 0x3
		  || header->layer == 4 || emphasis == 2)
		return FALSE;

	if (header->version == 0)
		header->bitrate = bitrates[0][header->layer-1][bitrate_idx];
	else
		header->bitrate = bitrates[1][header->layer-1][bitrate_idx];
	header->samplerate = s_freq[header->version][samplerate_idx];
	header->samples_per_frame = samples_per_frame[header->layer-1][header->version];
	header->padding = (buffer[2] >> 1) & 0x01;
	header->extension = buffer[2] & 0x01;
	header->channels = mode /*(mode == 0x3) ? 1 : 2*/;
	header->frame_length = mp3_frame_length(header);
	header->number_of_frames = 0;

	return TRUE;
}

static bool mp3_check_vbr(byte* buffer, mp3_header *header) 
{
	byte* data = buffer;
	int offset;

	/* check for VBR tag */
	/* Tag offset varies (but FhG VBRI is always MPEG1 Layer III 160 kbps stereo) */
	if (header->version == 0) {
		if (header->channels == 1)
			offset = 21;
		else
			offset = 36;
	} else {
		if (header->channels == 1)
			offset = 13;
		else
			offset = 21;
	}

	if (!strncmp("Xing", data + offset, 4)) {
		header->number_of_frames = (data[offset + 8]<<24) + (data[offset + 9]<<16) + (data[offset + 10]<<8) + data[offset + 11];
		return TRUE;
	}

	if (!strncmp("VBRI", data + 36, 4)) {
		header->number_of_frames = (data[offset + 14]<<24) + (data[offset + 15]<<16) + (data[offset + 16]<<8) + data[offset + 17];
		return TRUE;
	}

	return FALSE;
}

/* Calculate the expected length of the next frame, or return 0 if we don't know how */
static dword mp3_frame_length(mp3_header* header) 
{
	if (!header->bitrate)
		return 0;

	if (header->layer == 1) {
		return (12000 * header->bitrate / header->samplerate + header->padding) * 4;
	} else if (header->layer ==3 && header->version > 0) { 
		return 72000 * header->bitrate / header->samplerate + header->padding;
	}

	return 144000 * header->bitrate / header->samplerate + header->padding;
}

bool parse_blob_ID3(ID3 *info, byte *id3v1_blob, byte *id3v2_blob, size_t id3v1_blob_length, size_t id3v2_blob_length, bool read_art) 
{
	int rc1 = ID3_ERR_NO_TAG, rc2 = ID3_ERR_NO_TAG;

	// Sometimes people encode both headers, so we test for both

	// Why 127 instead of 128? The file may only have a tag in it
	if (id3v1_blob_length > 127) 
		rc1 = get_id3v1_tag(info, id3v1_blob, id3v1_blob_length);

	rc2 = get_id3v2_tag(info, id3v2_blob, id3v2_blob_length, read_art);

	//if (info->mask & KEEP_BLOB || info->mask & SIGNATURE_TAG)
	//  md5_signature(blob, blob_length, info->signature);

	return !rc1 || !rc2;
}

int destroy_ID3(ID3 *info) 
{
  if (info && info->buffer)
    dlfree(info->buffer);
  if (info && info->picture)
    dlfree(info->picture);
  //if (info->mask & KEEP_BLOB)
  //  munmap(info->data, info->data_size);
  if (info && info->allocated)
    dlfree(info);

  dlmalloc_trim(0);

  return 0; //Eventually we should do more here
}

ID3 * create_ID3(ID3 *info) 
{
  //At the moment ID3 is just a structure
  if (info) 
  {
    id3flags mask= info->mask;
    void *func= info->processor;
    size_t size= info->size;
    char *buffer= info->buffer;
    int user_memory= info->user_memory;
	if (info->picture)
		dlfree(info->picture);
    memset(info, 0, sizeof(ID3));
    info->size= size;
    info->buffer= buffer;
    info->mask= mask;
    info->user_memory= user_memory;
    info->allocated= 1;
    info->processor= func ? func : select_tag;
    info->ptr= info->buffer;
  } 
  else 
  {
    info= (ID3 *)dlmalloc(sizeof(ID3));
    if (info) 
      memset(info, 0, sizeof(ID3));
    else {
      //assert(0);
      return NULL;
    }

    info->allocated= 1;
    info->processor= select_tag;
    info->mask= ALL_TAG;
  }

  return info;
}
